from mrp import life_cycle
from mrp import util
from mrp.process_def import ProcDef
from mrp.runtime.base import BaseLauncher, BaseRuntime
import aiodocker
import asyncio
import contextlib
import docker
import io
import json
import os
import pathlib
import pwd
import typing


class Launcher(BaseLauncher):
    def __init__(
        self,
        image,
        mount: typing.List[str],
        name: str,
        proc_def: ProcDef,
        **kwargs,
    ):
        self.image = image
        self.mount = mount
        self.name = name
        self.proc_def = proc_def
        self.kwargs = kwargs

    async def run(self):
        container = f"mrp_{self.name}"

        # Environmental variables.
        env = util.common_env(self.proc_def)
        for kv in self.kwargs.pop("Env", []):
            if "=" in kv:
                k, v = kv.split("=", 1)
                env[k] = v
            else:
                del env[kv]
        env.update(self.proc_def.env)

        # Docker labels.
        labels = {
            "mrp.name": self.name,
            "mrp.container": container,
        }
        labels.update(self.kwargs.pop("labels", {}))

        # Mount volumes.
        for f in [
            "group",
            "gshadow",
            "inputrc",
            "localtime",
            "passwd",
            "shadow",
            "subgid",
            "subuid",
            "sudoers",
        ]:
            self.mount.append(f"/etc/{f}:/etc/{f}:ro")

        self.mount.append("/var/lib/sss:/var/lib/sss:ro")

        id_info = pwd.getpwuid(os.getuid())

        # Compose all the arguments.
        run_kwargs = {
            "Image": self.image,
            "Env": ["=".join(kv) for kv in env.items()],
            "Labels": labels,
            "User": f"{id_info.pw_uid}:{id_info.pw_gid}",
            "HostConfig": {
                "Privileged": True,
                "NetworkMode": "host",
                "PidMode": "host",
                "IpcMode": "host",
                "Binds": self.mount,
                "GroupAdd": [str(i) for i in os.getgroups()],
            },
        }
        util.nested_dict_update(run_kwargs, self.kwargs)

        docker = aiodocker.Docker()

        try:
            self.proc = await docker.containers.create_or_replace(container, run_kwargs)
            await self.proc.start()
            proc_info = await self.proc.show()
        except Exception as e:
            life_cycle.set_state(
                self.name, life_cycle.State.STOPPED, return_code=-1, error_info=str(e)
            )
            await docker.close()
            raise RuntimeError(
                f"Failed to start docker process: name={self.name} reason={e}"
            )

        self.proc_pid = proc_info["State"]["Pid"]
        life_cycle.set_state(self.name, life_cycle.State.STARTED)

        async def log_pipe(logger, pipe):
            async for line in pipe:
                logger(line)

        self.down_task = asyncio.create_task(self.down_watcher(self.handle_down))
        try:
            await asyncio.gather(
                log_pipe(util.stdout_logger(), self.proc.log(stdout=True, follow=True)),
                log_pipe(util.stderr_logger(), self.proc.log(stderr=True, follow=True)),
                self.log_psutil(),
                self.death_handler(),
                self.down_task,
            )
        except asyncio.exceptions.CancelledError:
            # death_handler cancelled down listener.
            pass
        await docker.close()

    def get_pid(self):
        return self.proc_pid

    async def death_handler(self):
        result = await self.proc.wait()
        life_cycle.set_state(
            self.name, life_cycle.State.STOPPED, return_code=result["StatusCode"]
        )
        # TODO(lshamis): Restart policy goes here.

        # Release the down listener.
        self.down_task.cancel()

    async def handle_down(self):
        await self.proc.stop()
        with contextlib.suppress(asyncio.TimeoutError):
            await self.proc.wait(timeout=3.0)
        await self.proc.delete(force=True)


class Docker(BaseRuntime):
    def __init__(
        self,
        image=None,
        dockerfile=None,
        mount=None,
        build_kwargs=None,
        run_kwargs=None,
    ):
        mount = mount or []
        build_kwargs = build_kwargs or {}
        run_kwargs = run_kwargs or {}

        if bool(image) == bool(dockerfile):
            raise ValueError(
                "Docker process must define exactly one of image or dockerfile"
            )
        self.image = image
        self.dockerfile = dockerfile
        self.mount = mount
        self.build_kwargs = build_kwargs
        self.run_kwargs = run_kwargs

    def asdict(self, root: pathlib.Path):
        ret = {}
        if self.image:
            ret["image"] = self.image
        if self.dockerfile:
            ret["dockerfile"] = self.dockerfile
        if self.mount:
            ret["mount"] = self.mount
        if self.build_kwargs:
            ret["build_kwargs"] = self.build_kwargs
        if self.run_kwargs:
            ret["run_kwargs"] = self.run_kwargs
        return ret

    def _build(self, name: str, proc_def: ProcDef, cache: bool, verbose: bool):
        docker_api = docker.from_env()
        docker_api.lowlevel = docker.APIClient()

        if self.dockerfile:
            self.image = f"mrp/{name}"

            dockerfile_path = os.path.join(
                os.path.dirname(proc_def.rule_file), self.dockerfile
            )

            build_kwargs = {
                "tag": self.image,
                "path": proc_def.root,
                "dockerfile": dockerfile_path,
                "rm": True,
                "nocache": not cache,
            }
            build_kwargs.update(self.build_kwargs)

            for line in docker_api.lowlevel.build(**build_kwargs):
                lineinfo = json.loads(line.decode())
                if verbose and "stream" in lineinfo:
                    print(lineinfo["stream"].strip())
                elif "errorDetail" in lineinfo:
                    raise RuntimeError(json.dumps(lineinfo["errorDetail"], indent=2))
        else:
            should_pull = not cache
            if cache:
                try:
                    docker_api.images.get(self.image)
                except docker.errors.ImageNotFound:
                    should_pull = True

            if should_pull:
                try:
                    for line in docker_api.lowlevel.pull(self.image, stream=True):
                        lineinfo = json.loads(line.decode())
                        if verbose and "status" in lineinfo:
                            print(lineinfo["status"].strip())
                except docker.errors.NotFound as e:
                    raise RuntimeError(e)

        uses_ldap = util.is_ldap_user()

        mount_map = {}
        for mnt in self.mount:
            parts = mnt.split(":")
            if len(parts) == 3:
                mount_map[parts[0]] = (parts[1], parts[2])
            elif len(parts) == 2:
                mount_map[parts[0]] = (parts[1], "")
            else:
                raise ValueError(f"Invalid mount: {mnt}")

        nfs_mounts = []
        for host, (container, _) in mount_map.items():
            try:
                os.makedirs(host)
            except FileExistsError:
                pass
            except FileNotFoundError:
                pass

            nfs_root = util.nfs_root(host)
            if nfs_root and nfs_root != host:
                nfs_mounts.append(
                    {
                        "host_nfs_root": nfs_root,
                        "container_nfs_root": f"/mrp_nfs/{util.random_string()}",
                        "host": host,
                        "container": container,
                    }
                )

        for nfs_mount in nfs_mounts:
            if nfs_mount["host"] in mount_map:
                del mount_map[nfs_mount["host"]]

        if uses_ldap or nfs_mounts:
            dockerfile = [f"FROM {self.image}"]
            if uses_ldap:
                # We try to inject sssd to allow the container to communicate with the ldap server.
                # We assume the docker container provides apt. If there is no apt, we skip this step. The user id and group ids will still correctly match the active user, but the user and group names will not be fetchable.
                apt_install_sssd_cmd = " && ".join(
                    [
                        "apt update",
                        "DEBIAN_FRONTEND=noninteractive apt install -y sssd",
                        "rm -rf /var/lib/apt/lists/*",
                    ]
                )
                try_apt_install_sssd_cmd = (
                    f"if [ $(which apt) ] ; then $({apt_install_sssd_cmd}) ; fi"
                )
                dockerfile.append(f"RUN {try_apt_install_sssd_cmd}")
            for nfs_mount in nfs_mounts:
                relpath = os.path.relpath(nfs_mount["host"], nfs_mount["host_nfs_root"])
                container_fullpath = os.path.join(
                    nfs_mount["container_nfs_root"], relpath
                )
                mkdir = f"mkdir -p {os.path.dirname(nfs_mount['container'])}"
                link = f"ln -s {container_fullpath} {nfs_mount['container']}"
                dockerfile.append(f"RUN {mkdir} && {link}")

                mount_map[nfs_mount["host_nfs_root"]] = (
                    nfs_mount["container_nfs_root"],
                    "",
                )

            self.image = f"mrp/{name}:netext"

            for line in docker_api.lowlevel.build(
                fileobj=io.BytesIO("\n".join(dockerfile).encode("utf-8")),
                rm=True,
                tag=self.image,
                labels={
                    "mrp.type": "netext",
                    "mrp.tmp": "1",
                },
            ):
                lineinfo = json.loads(line.decode())
                if verbose and "stream" in lineinfo:
                    print(lineinfo["stream"].strip())
                elif "errorDetail" in lineinfo:
                    raise RuntimeError(json.dumps(lineinfo["errorDetail"], indent=2))

        self.mount = [
            f"{host_path}:{container_path}:{options}"
            for host_path, (container_path, options) in mount_map.items()
        ]

    def _launcher(self, name: str, proc_def: ProcDef):
        return Launcher(self.image, self.mount, name, proc_def, **self.run_kwargs)



import a0
import collections.abc
import contextlib
import glob
import os
import pwd
import random
import shlex
import string
import subprocess


def common_env(proc_def):
    return dict(
        MRP_NAME=proc_def.name,
        A0_TOPIC=proc_def.name,
        PYTHONUNBUFFERED="1",
    )


@contextlib.contextmanager
def common_env_context(proc_def):
    old_environ = dict(os.environ)
    os.environ.update(common_env(proc_def))
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


def _make_plaintext_logger(logger_callback):
    logger = a0.Logger(a0.env.topic())

    def write(msg):
        pkt = a0.Packet([("content-type", "text/plain")], msg)
        logger_callback(logger, pkt)

    return write


def stdout_logger():
    return _make_plaintext_logger(lambda logger, pkt: logger.info(pkt))


def stderr_logger():
    return _make_plaintext_logger(lambda logger, pkt: logger.err(pkt))


def pid_children(pid):
    for fname in glob.glob(f"/proc/{pid}/task/*/children"):
        with open(fname) as file:
            for line in file:
                yield int(line)


def is_ldap_user():
    user = pwd.getpwuid(os.getuid()).pw_name
    for line in open("/etc/passwd"):
        if line.startswith(f"{user}:"):
            return False
    return True


def nfs_root(path):
    result = subprocess.run(["findmnt", "-T", path], capture_output=True)
    lines = result.stdout.decode().split("\n")
    if len(lines) < 2:
        return None
    properties = dict(zip(lines[0].split(), lines[1].split()))
    if properties["FSTYPE"] != "nfs":
        return None
    return properties["TARGET"]


def random_string(alphabet=string.ascii_lowercase, length=16):
    return "".join(random.choices(alphabet, k=length))


# https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
def nested_dict_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = nested_dict_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class NoEscape:
    def __init__(self, val):
        self.value = val


def shell_join(items):
    """Modified version of shlex.join that allows for non-escaped segments."""
    return " ".join(
        item.value if type(item) == NoEscape else shlex.quote(item) for item in items
    )



from mrp import life_cycle
from mrp.process_def import ProcDef
import a0
import asyncio
import contextlib
import json
import pathlib
import psutil


# psutil as_dict() does not produce a json-serializable dict.
# https://github.com/giampaolo/psutil/issues/967
#
# When converted to a dict, fields like memory_info generate
#     [11984896, 31031296, ...]
# instead of
#     {"rss": 11984896, "vms": 31031296, ...}
# Losing field names.
#
# We cannot use a custom JSONEncoder, since psutil objects, like memory_info,
# inherit from tuples.
#
# This is a workaround.
def _walk_asdict(obj):
    if hasattr(obj, "_asdict"):
        return _walk_asdict(obj._asdict())
    if isinstance(obj, dict):
        return {k: _walk_asdict(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [_walk_asdict(v) for v in obj]
    return obj


class BaseLauncher:
    def __init__(self):
        pass

    async def run(self, name: str, proc_def: ProcDef):
        raise NotImplementedError("Launcher hasn't implemented run!")

    def get_pid(self):
        raise NotImplementedError("Launcher hasn't implemented get_pid!")

    async def down_watcher(self, ondown):
        async for proc_info in life_cycle.aio_proc_info_watcher(self.name):
            if proc_info.ask == life_cycle.Ask.DOWN:
                await ondown()
                break

    async def log_psutil(self):
        down_requested_event = asyncio.Event()

        async def ondown():
            down_requested_event.set()

        asyncio.ensure_future(self.down_watcher(ondown))

        out = a0.Publisher(f"mrp/psutil/{self.name}")
        while True:
            with contextlib.suppress(asyncio.TimeoutError):
                # TODO(lshamis): Make polling interval configurable.
                await asyncio.wait_for(down_requested_event.wait(), 1.0)
            if down_requested_event.is_set():
                break
            pid = self.get_pid()
            if not pid:
                # Note: Likely being restarted.
                continue

            try:
                proc = psutil.Process(pid)
                pkt = a0.Packet(
                    [("content-type", "application/json")],
                    json.dumps(_walk_asdict(proc.as_dict())),
                )
                out.pub(pkt)
            except psutil.NoSuchProcess:
                pass


class BaseRuntime:
    def __init__(self):
        pass

    def asdict(self, root: pathlib.Path):
        raise NotImplementedError("Runtime hasn't implemented asdict!")

    def _build(self, name: str, proc_def: ProcDef, cache: bool, verbose: bool):
        raise NotImplementedError("Runtime hasn't implemented build!")

    def _launcher(self, name: str, proc_def: ProcDef) -> BaseLauncher:
        raise NotImplementedError("Runtime hasn't implemented a launcher!")



from mrp import life_cycle
from mrp import process_def
from mrp import util
from mrp.cmd import _autocomplete
import a0
import asyncio
import click
import contextlib
import json
import os
import threading
import traceback
import types
import typing


def transitive_closure(proc_names):
    all_proc_names = set()
    fringe = list(proc_names)
    while fringe:
        proc_name = fringe.pop()
        if proc_name in all_proc_names:
            continue
        try:
            fringe.extend(process_def.defined_processes[proc_name].deps)
        except KeyError:
            raise ValueError(f"Unknown process: {proc_name}")
        all_proc_names.add(proc_name)
    return all_proc_names


def get_proc_names(proc_names, include_deps):
    if not proc_names:
        return process_def.defined_processes.keys()

    if include_deps:
        proc_names = transitive_closure(proc_names)
    unknown_proc_names = [
        proc_name
        for proc_name in proc_names
        if proc_name not in process_def.defined_processes
    ]
    if unknown_proc_names:
        raise ValueError(f"Unknown proc_names: {', '.join(unknown_proc_names)}")
    if not proc_names:
        raise ValueError("No proc_names found")
    return proc_names


def down_existing(names: typing.List[str], force: bool):
    def find_active_proc(system_state):
        return [
            name
            for name in names
            if name in system_state.procs
            and system_state.procs[name].state != life_cycle.State.STOPPED
        ]

    active_proc = find_active_proc(life_cycle.system_state())
    if not active_proc:
        return

    if not force:
        raise RuntimeError(
            f"Conflicting processes already running: {', '.join(active_proc)}"
        )

    for name in active_proc:
        life_cycle.set_ask(name, life_cycle.Ask.DOWN)

    ns = types.SimpleNamespace()
    ns.cv = threading.Condition()
    ns.sat = False

    def callback(system_state):
        if not find_active_proc(system_state):
            with ns.cv:
                ns.sat = True
                ns.cv.notify()

    watcher = life_cycle.system_state_watcher(callback)  # noqa: F841

    with ns.cv:
        success = ns.cv.wait_for(lambda: ns.sat, timeout=3.0)

    if not success:
        raise RuntimeError("Existing processes did not down in a timely manner.")


@click.command()
@click.argument("procs", nargs=-1, shell_complete=_autocomplete.defined_processes)
@click.option("-v/-q", "--verbose/--quiet", is_flag=True, default=True)
@click.option("--deps/--nodeps", is_flag=True, default=True)
@click.option("--build/--nobuild", is_flag=True, default=True)
@click.option("--cache/--nocache", is_flag=True, default=True)
@click.option("--run/--norun", is_flag=True, default=True)
@click.option("-f", "--force/--noforce", is_flag=True, default=False)
@click.option("--reset_logs", is_flag=True, default=False)
def cli(
    *cmd_procs,
    procs=None,
    verbose=True,
    deps=True,
    build=True,
    cache=True,
    run=True,
    force=False,
    reset_logs=False,
):
    procs = procs or []

    # Support procs as *args when using cmd syntax.
    procs += cmd_procs

    names = get_proc_names(procs, deps)
    names = [name for name in names if process_def.defined_processes[name].runtime]
    if not names:
        raise ValueError("No processes found")

    down_existing(names, force)

    if reset_logs:
        for name in names:
            a0.File.remove(f"{name}.log.a0")

    if build:
        for name in names:
            proc_def = process_def.defined_processes[name]
            click.echo(f"building {name}...")
            proc_def.runtime._build(name, proc_def, cache, verbose)
            click.echo(f"built {name}\n")

    if run:
        for name in names:
            click.echo(f"running {name}...")
            life_cycle.set_ask(name, life_cycle.Ask.UP)
            life_cycle.set_state(name, life_cycle.State.STARTING)

            if os.fork() != 0:
                continue

            os.chdir("/")
            os.setsid()
            os.umask(0)

            if os.fork() != 0:
                os._exit(0)  # use this instead of sys.exit in child process

            proc_def = process_def.defined_processes[name]

            # Set up configuration.
            with util.common_env_context(proc_def):
                a0.Cfg(a0.env.topic()).write(json.dumps(proc_def.cfg))

                with open(f"/tmp/mrp_{name}.log", "w", buffering=1) as logfile:
                    with contextlib.redirect_stdout(
                        logfile
                    ), contextlib.redirect_stderr(logfile):
                        click.echo(f"-- Process start time {a0.TimeWall.now()}")
                        life_cycle.set_launcher_running(name, True)
                        try:
                            asyncio.run(
                                proc_def.runtime._launcher(name, proc_def).run()
                            )
                        except BaseException as e:
                            click.echo(f"FATAL: {e}")
                            traceback.print_exc()
                        life_cycle.set_launcher_running(name, False)
                        os._exit(0)  # use this instead of sys.exit in child process



import a0
import dataclasses
import enum
import json
import typing
import types

_TOPIC = "mrp/state"
_CFG = a0.Cfg(_TOPIC)


class Ask(enum.Enum):
    NONE = "NONE"
    UP = "UP"
    DOWN = "DOWN"


class State(enum.Enum):
    UNKNOWN = "UNKNOWN"
    STARTING = "STARTING"
    STARTED = "STARTED"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"


@dataclasses.dataclass(frozen=True)
class ProcInfo:
    ask: Ask
    state: State
    return_code: int
    launcher_running: bool
    error_info: str

    def asdict(self):
        return dict(
            ask=self.ask.value,
            state=self.state.value,
            return_code=self.return_code,
            launcher_running=self.launcher_running,
            error_info=self.error_info,
        )

    @classmethod
    def fromdict(cls, dict_):
        return cls(
            ask=Ask(dict_.get("ask", Ask.NONE)),
            state=State(dict_.get("state", State.UNKNOWN)),
            return_code=dict_.get("return_code", 0),
            launcher_running=dict_.get("launcher_running", False),
            error_info=dict_.get("error_info", ""),
        )


@dataclasses.dataclass
class SystemState:
    procs: typing.Mapping[str, ProcInfo]

    def asdict(self):
        return {"procs": {k: v.asdict() for k, v in self.procs.items()}}

    @classmethod
    def fromdict(cls, dict_):
        return SystemState(
            procs={k: ProcInfo.fromdict(v) for k, v in dict_.get("procs", {}).items()}
        )


def _ensure_setup() -> None:
    _CFG.write_if_empty(json.dumps(SystemState(procs={}).asdict()))


def system_state() -> SystemState:
    _ensure_setup()
    return SystemState.fromdict(json.loads(_CFG.read().payload))


def proc_info(proc_name) -> ProcInfo:
    return system_state().procs[proc_name]


def system_state_watcher(callback) -> a0.CfgWatcher:
    _ensure_setup()

    def callback_wrapper(pkt):
        callback(SystemState.fromdict(json.loads(pkt.payload)))

    return a0.CfgWatcher(_TOPIC, callback_wrapper)


async def aio_system_state_watcher():
    _ensure_setup()
    async for pkt in a0.aio_cfg(_TOPIC):
        yield SystemState.fromdict(json.loads(pkt.payload))


def proc_info_watcher(proc_name, callback) -> a0.CfgWatcher:
    ns = types.SimpleNamespace()
    ns.last_proc_info = None

    def callback_wrapper(system_state):
        if ns.last_proc_info != system_state.procs.get(proc_name):
            ns.last_proc_info = system_state.procs[proc_name]
            callback(system_state.procs[proc_name])

    return system_state_watcher(callback_wrapper)


async def aio_proc_info_watcher(proc_name):
    ns = types.SimpleNamespace()
    ns.last_proc_info = None

    async for system_state in aio_system_state_watcher():
        if ns.last_proc_info != system_state.procs.get(proc_name):
            ns.last_proc_info = system_state.procs[proc_name]
            yield system_state.procs[proc_name]


def set_ask(proc_name, ask):
    _CFG.mergepatch({"procs": {proc_name: {"ask": ask.value}}})


def set_state(proc_name, state, return_code=0, error_info=""):
    _CFG.mergepatch(
        {
            "procs": {
                proc_name: {
                    "state": state.value,
                    "return_code": return_code,
                    "error_info": error_info,
                }
            }
        }
    )


def set_launcher_running(proc_name, launcher_running):
    _CFG.mergepatch({"procs": {proc_name: {"launcher_running": launcher_running}}})



import dataclasses
import inspect
import os
import pathlib
import typing

if typing.TYPE_CHECKING:
    from mrp.runtime.base import BaseRuntime


@dataclasses.dataclass
class ProcDef:
    name: str
    root: pathlib.Path
    rule_file: pathlib.Path
    runtime: "BaseRuntime"
    cfg: dict
    deps: typing.List[str]
    env: dict

    def asdict(self):
        return {
            "name": self.name,
            "root": str(self.root),
            "rule_file": str(self.rule_file),
            "runtime": self.runtime.asdict(self.root),
            "cfg": self.cfg,
            "deps": self.deps,
            "env": self.env,
        }


defined_processes: typing.Dict[str, ProcDef] = {}


def process(
    name: str,
    root: str = None,
    runtime: "BaseRuntime" = None,
    cfg: typing.Optional[dict] = None,
    deps: typing.Optional[typing.List[str]] = None,
    env: typing.Optional[dict] = None,
) -> ProcDef:
    deps = deps or []
    cfg = cfg or {}
    env = env or {}

    if name in defined_processes:
        raise ValueError(f"mrp.process(name={name}) defined multiple times.")

    rule_file = os.path.abspath(inspect.stack()[1].filename)
    rule_dir = os.path.dirname(rule_file)

    if root:
        root = os.path.normpath(os.path.join(rule_dir, root))
    else:
        root = rule_dir

    # Validate env is dict[str, str]
    for k, v in env.items():
        if [type(k), type(v)] != [str, str]:
            raise ValueError(
                f"mrp.process(name={name}) invalid. env is not dict[str, str]"
            )

    defined_processes[name] = ProcDef(
        name=name,
        root=root,
        runtime=runtime,
        cfg=cfg,
        deps=deps,
        rule_file=rule_file,
        env=env,
    )

    return defined_processes[name]



from mrp import util
from mrp.cmd import _autocomplete
from mrp import process_def
import a0
import click
import random
import signal
import sys


@click.command()
@click.argument(
    "procs",
    nargs=-1,
    # Logs connects to live feeds unless --old is set.
    # This autocomplete checks for the old flag.
    #   If --old is set, suggest any defined process that has logs.
    #   Otherwise, suggest running processes.
    shell_complete=_autocomplete.conditional(
        lambda ctx, unused_param, unused_incomplete: ctx.params["old"],
        _autocomplete.intersection(
            _autocomplete.alephzero_topics(protocol="log"),
            _autocomplete.defined_processes,
        ),
        _autocomplete.running_processes,
    ),
)
@click.option("-o", "--old", is_flag=True, default=False)
def cli(*cmd_procs, procs=None, old=False):
    procs = procs or []

    # Support procs as *args when using cmd syntax.
    procs += cmd_procs

    # Find all defined processes.
    display_procs = process_def.defined_processes.items()
    # Filter out processes that have no runtime defined.
    # These processes were meant to chain or combine other processes, but haven't
    # gotten much use yet. Do we want to keep them?
    display_procs = {name: def_ for name, def_ in display_procs if def_.runtime}

    # If processes have been specified, filter out the ones that aren't requested.
    if procs:
        display_procs = {
            name: def_ for name, def_ in display_procs.items() if name in procs
        }

    # Fail if no processes are left.
    if not display_procs:
        raise ValueError("No processes found to log")

    # Give each process a random color.
    colors = [
        "red",
        "green",
        "yellow",
        "blue",
        "magenta",
        "cyan",
    ]
    random.shuffle(colors)

    # There will be a left hand column with the process name.
    # Find a common width for the name column.
    width = max(len(name) for name in display_procs)

    log_listeners = []

    def make_listener(i, name, def_):
        # Cache the left hand column.
        prefix = f"{name}" + " " * (width - len(name))
        msg_tmpl = f"{prefix} | {{msg}}"

        # On message received, print it to stdout.
        def callback(pkt):
            click.secho(msg_tmpl.format(msg=pkt.payload), fg=colors[i % len(colors)])

        # Create the listener.
        with util.common_env_context(def_):
            log_listeners.append(
                a0.LogListener(
                    name,
                    # TODO(lshamis): Make a flag for log level.
                    a0.LogLevel.DBG,
                    a0.INIT_OLDEST if old else a0.INIT_AWAIT_NEW,
                    a0.ITER_NEXT,
                    callback,
                )
            )

    # Make a log listener for each process.
    for i, (name, def_) in enumerate(display_procs.items()):
        make_listener(i, name, def_)

    # Block until ctrl-c is pressed.
    def onsignal(signum, frame):
        sys.exit(0)

    signal.signal(signal.SIGINT, onsignal)
    signal.signal(signal.SIGTERM, onsignal)
    signal.pause()



import mrp
from mrp import life_cycle
from mrp import process_def
from mrp.cmd import _autocomplete
import click


@click.command()
@click.option("--all", is_flag=True)
@click.option("--wait/--nowait", is_flag=True)
@click.argument("procs", nargs=-1, shell_complete=_autocomplete.defined_processes)
def cli(*cmd_procs, procs=None, all=False, wait=True):
    procs = procs or []

    # Get all MRP procs running in the system
    running_procs = life_cycle.system_state().procs.keys()
    down_procs = set(running_procs)

    if all:  # system-wide down
        assert not procs, "Specifying processes is not supported with the flag '--all'."
        assert not cmd_procs, "Specifying processes is not supported when all=True."

    else:  #  local down (only processes defined within the current msetup.py)
        defined_procs = process_def.defined_processes.keys()
        down_procs = down_procs & set(defined_procs)

        # Support procs as *args when using cmd syntax.
        procs += cmd_procs
        if procs:
            down_procs = down_procs & set(procs)

    for proc in down_procs:
        click.echo(f"stopping {proc}...")
        life_cycle.set_ask(proc, life_cycle.Ask.DOWN)

    if wait and down_procs:
        mrp.cmd.wait(*down_procs)



from mrp import life_cycle
from mrp import util
from mrp.process_def import ProcDef
from mrp.runtime.base import BaseLauncher, BaseRuntime
import asyncio
import dataclasses
import filecmp
import json
import os
import pathlib
import shutil
import signal
import subprocess
import typing
import yaml as pyyaml


@dataclasses.dataclass
class CondaEnv:
    channels: typing.List[str]
    dependencies: typing.List[typing.Union[str, dict]]

    @staticmethod
    def load(flo) -> "CondaEnv":
        """Load a conda env definition from a yaml flo."""
        result = pyyaml.safe_load(flo)
        return CondaEnv(result.get("channels", []), result.get("dependencies", []))

    @staticmethod
    def from_named_env(name) -> "CondaEnv":
        """Load a conda env definition from a named environment."""
        return CondaEnv.load(
            subprocess.check_output(["conda", "env", "export", "-n", name])
        )

    @staticmethod
    def merge(lhs: "CondaEnv", rhs: "CondaEnv") -> "CondaEnv":
        """Merge two conda environments."""
        channels = sorted(set(lhs.channels + rhs.channels))

        deps = []
        pip_deps = []
        for dep in lhs.dependencies + rhs.dependencies:
            if type(dep) == dict and dep.get("pip"):
                pip_deps.extend(dep["pip"])
            else:
                deps.append(dep)

        deps = sorted(set(deps))

        if pip_deps:
            pip_deps = sorted(set(pip_deps))
            deps.append({"pip": pip_deps})

        return CondaEnv(channels, deps)

    def fix_empty(self):
        """Fix empty dependencies.

        Conda will not create an environment if the dependencies list is empty.
        """
        if not self.dependencies:
            self.dependencies = ["python"]

    def fix_pip(self):
        """Fix pip dependencies.

        Conda special cases pip dependencies with the form:
        {"pip": ["pkg_a", "pkg_b"]}

        But still requires a top-level pip dependency.
        This method adds a top-level pip dependency if not present.
        """
        if any(dep is str and dep.startswith("pip") for dep in self.dependencies):
            return
        for dep in self.dependencies:
            if type(dep) == dict and dep.get("pip"):
                self.dependencies.append("pip")
                return


class Launcher(BaseLauncher):
    def __init__(
        self,
        run_command: typing.List[str],
        name: str,
        env_name: str,
        proc_def: ProcDef,
    ):
        self.run_command = run_command
        self.name = name
        self.env_name = env_name
        self.proc_def = proc_def

    async def envvar_for_conda(self) -> dict:
        """Detect the envvar set by conda for our env."""
        # We grab the conda env variables separate from executing the run
        # command to simplify detecting pid and removing some race conditions.
        subprocess_env = os.environ.copy()
        subprocess_env.update(self.proc_def.env)
        envvar_file = f"/tmp/mrp_conda_{self.name}.env"
        envvar_info = await asyncio.create_subprocess_shell(
            f"""
                eval "$(conda shell.bash hook)"
                conda activate {self.env_name}
                cp -f /proc/self/environ {envvar_file}
            """,
            stdout=asyncio.subprocess.PIPE,
            executable="/bin/bash",
            cwd=self.proc_def.root,
            env=subprocess_env,
        )

        await envvar_info.wait()
        lines = open(envvar_file).read().split("\0")
        return dict(line.split("=", 1) for line in lines if "=" in line)

    async def run_cmd_with_conda_envvar(self, conda_envvar):
        """Run the command with the conda envvar."""
        self.proc = await asyncio.create_subprocess_shell(
            util.shell_join(self.run_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self.proc_def.root,
            env=conda_envvar,
            start_new_session=True,
        )
        try:
            self.proc_pgrp = os.getpgid(self.proc.pid)
            return True
        except Exception:
            return False

    async def gather_cmd_outputs(self):
        """Track the command.

        This includes:
        - piping stdout and stderr to the alephzero logger.
        - logging the processes psutil stats.
        - watching for process death.
        - watching for user down request.
        """

        async def log_pipe(logger, pipe):
            while True:
                try:
                    async for line in pipe:
                        logger(line)
                except ValueError:
                    # TODO(lshamis): Can we grab the line in chucks?
                    logger("<[MRP] line length exceeded. skipping>")
                    continue
                else:
                    break

        self.down_task = asyncio.create_task(self.down_watcher(self.handle_down))
        try:
            await asyncio.gather(
                log_pipe(util.stdout_logger(), self.proc.stdout),
                log_pipe(util.stderr_logger(), self.proc.stderr),
                self.log_psutil(),
                self.death_handler(),
                self.down_task,
            )
        except asyncio.exceptions.CancelledError:
            # death_handler cancelled down listener.
            pass

    async def run(self):
        """Run the command."""
        conda_envvar = await self.envvar_for_conda()
        if await self.run_cmd_with_conda_envvar(conda_envvar):
            life_cycle.set_state(self.name, life_cycle.State.STARTED)
            await self.gather_cmd_outputs()
        life_cycle.set_state(
            self.name, life_cycle.State.STOPPED, return_code=self.proc.returncode
        )

    def get_pid(self):
        return self.proc.pid

    async def death_handler(self):
        """Watch for process death."""
        await self.proc.wait()
        # TODO(lshamis): Restart policy goes here.

        # Release the down listener.
        self.down_task.cancel()

    async def handle_down(self):
        """Handle user down request."""
        try:
            # If still running...
            if self.proc.returncode is None:
                life_cycle.set_state(self.name, life_cycle.State.STOPPING)
                # Send a gentle SIGTERM to let the process know it's time to die.
                self.proc.send_signal(signal.SIGTERM)

                try:
                    # Wait for the process to die.
                    await asyncio.wait_for(self.proc.wait(), timeout=3.0)
                except asyncio.TimeoutError:
                    pass

            # Send a SIGKILL to force the process group to kill stray processes.
            os.killpg(self.proc_pgrp, signal.SIGKILL)
        except:
            pass


class Conda(BaseRuntime):
    class SharedEnv:
        def __init__(
            self,
            name,
            use_named_env=None,
            copy_named_env=None,
            yaml=None,
            channels=[],
            dependencies=[],
            use_mamba=None,
            setup_commands=[],
        ):
            if use_named_env and any([copy_named_env, yaml, channels, dependencies]):
                raise ValueError(
                    "'use_named_env' cannot be mixed with other Conda environment commands."
                )

            self.name = name
            self.use_named_env = use_named_env
            self.copy_named_env = copy_named_env
            self.yaml = yaml
            self.conda_env = CondaEnv(channels[:], dependencies[:])
            self.use_mamba = (
                use_mamba if use_mamba is not None else bool(shutil.which("mamba"))
            )
            self.setup_commands = setup_commands
            self._built = False

            if use_named_env:
                self._validate_use_named_env()

        def _validate_use_named_env(self):
            """Validate that the named env exists."""
            result = subprocess.run(
                f"""
                    eval "$(conda shell.bash hook)"
                    conda activate {self.use_named_env}
                """,
                shell=True,
                executable="/bin/bash",
                stderr=subprocess.PIPE,
            )
            if result.returncode:
                raise RuntimeError(f"'use_named_env' not valid: {result.stderr}")

        def _generate_env_content(self, root: pathlib.Path) -> dict:
            """Generate the conda environment content.

            This includes the channels and dependencies obtained from:
            - explicit channels and dependencies.
            - querying a named env to duplicate.
            - given yaml file.
            """
            if self.copy_named_env:
                self.conda_env = CondaEnv.merge(
                    self.conda_env, CondaEnv.from_named_env(self.copy_named_env)
                )

            if self.yaml:
                yaml_path = os.path.join(root, self.yaml)
                self.conda_env = CondaEnv.merge(
                    self.conda_env, CondaEnv.load(open(yaml_path, "r"))
                )

            self.conda_env.fix_empty()
            self.conda_env.fix_pip()
            return {
                "channels": self.conda_env.channels,
                "dependencies": self.conda_env.dependencies,
            }

        def _env_name(self):
            """The target environment name."""
            return self.use_named_env or f"mrp_{self.name}"

        def _config_path(self):
            dirpath = os.path.expanduser(f"~/.config/mrp/conda/{self._env_name()}/")
            os.makedirs(dirpath, exist_ok=True)
            return dirpath

        def _yaml_path(self):
            return os.path.join(self._config_path(), "conda_env.yaml")

        def _setup_commands_snapshot_path(self):
            return os.path.join(self._config_path(), "setup_commands.snapshot")

        def _conda_history_snapshot_path(self):
            return os.path.join(self._config_path(), "history.snapshot")

        def _conda_history_path(self):
            info = json.loads(
                subprocess.check_output(
                    ["conda", "env", "export", "--json", "-n", self._env_name()]
                )
            )

            return os.path.join(info["prefix"], "conda-meta/history")

        def _cache_valid(self, env_content):
            try:
                # Check if the env description has changed.
                old_content = json.load(open(self._yaml_path()))
                old_content_compare_key = json.dumps(old_content, sort_keys=True)
                new_content_compare_key = json.dumps(env_content, sort_keys=True)
                if old_content_compare_key != new_content_compare_key:
                    print("detected change in environment definition.")
                    return False

                # Check if the setup commands have changed.
                if self.setup_commands != json.load(
                    open(self._setup_commands_snapshot_path())
                ):
                    print("detected change in setup commands.")
                    return False

                # Check if unmanaged commands have been executed.
                if not filecmp.cmp(
                    self._conda_history_path(), self._conda_history_snapshot_path()
                ):
                    print("detected change in conda environment.")
                    return False

                return True
            except Exception:
                return False

        def _create_env(self, root: pathlib.Path, cache: bool, verbose: bool):
            """Create the conda environment."""
            yaml_path = self._yaml_path()
            env_content = self._generate_env_content(root)
            env_content["name"] = self._env_name()

            if cache and self._cache_valid(env_content):
                return

            with open(yaml_path, "w") as env_fp:
                json.dump(env_content, env_fp, indent=2)

            print(
                f"creating conda env for {self._env_name()}. This will take a minute..."
            )

            update_bin = "mamba" if self.use_mamba else "conda"
            # https://github.com/conda/conda/issues/7279
            # Updating an existing environment does not remove old packages, even with --prune.
            subprocess.run(
                [update_bin, "env", "remove", "-n", self._env_name()],
                capture_output=not verbose,
            )
            result = subprocess.run(
                [update_bin, "env", "update", "--prune", "-f", yaml_path],
                capture_output=not verbose,
            )
            if result.returncode:
                raise RuntimeError(f"Failed to set up conda env: {result.stderr}")

            print(f"running setup commands for {self._env_name()}")

            setup_command = "\n".join(
                [util.shell_join(cmd) for cmd in self.setup_commands]
            )
            result = subprocess.run(
                f"""
                    eval "$(conda shell.bash hook)"
                    conda activate {self._env_name()}
                    cd {root}
                    {setup_command}
                """,
                shell=True,
                executable="/bin/bash",
                capture_output=not verbose,
            )
            if result.returncode:
                raise RuntimeError(f"Failed to set up conda env: {result.stderr}")

            # Snapshot successful build info.
            shutil.copy2(
                self._conda_history_path(), self._conda_history_snapshot_path()
            )
            json.dump(
                self.setup_commands, open(self._setup_commands_snapshot_path(), "w")
            )

        def _build(self, root: pathlib.Path, cache: bool, verbose: bool):
            if self._built:
                return
            if not self.use_named_env:
                self._create_env(root, cache, verbose)
            self._built = True

    def __init__(
        self,
        run_command,
        use_named_env=None,
        copy_named_env=None,
        shared_env=None,
        yaml=None,
        channels=[],
        dependencies=[],
        setup_commands=[],
        use_mamba=None,
    ):
        """Declare a conda runtime environment.

        Args:
            run_command: The command to run in the conda environment.
            use_named_env: The name of the conda environment to use. This is exclusive from other env definition arguments.
            copy_named_env: Create a new conda environment by duplicating a named environment.
            shared_env: A shared Conda.SharedEnv instance. This is exclusive from other env definition arguments.
            yaml: Create a new conda environment from the definition file.
            channels: Create a new conda environment using the given channels.
            dependencies: Create a new conda environment using the given dependencies.
            setup_commands: Commands to run during the build phase.
            use_mamba: Use mamba instead of conda. If not set, mamba will be autodetected.
        """
        if shared_env and any(
            [use_named_env, copy_named_env, yaml, channels, dependencies]
        ):
            raise ValueError(
                "'shared_env' cannot be mixed with other Conda environment commands."
            )

        self._env = shared_env
        self._is_personal_env = not shared_env
        if self._is_personal_env:
            self._env = Conda.SharedEnv(
                "__defer__",
                use_named_env,
                copy_named_env,
                yaml,
                channels,
                dependencies,
                use_mamba,
                setup_commands,
            )

        self.run_command = run_command

    def asdict(self, root: pathlib.Path):
        ret = {}
        if self._env.use_named_env:
            ret["use_named_env"] = self._env.use_named_env
        else:
            ret["env"] = self._env._generate_env_content(root)
        if self.run_command:
            ret["run_command"] = self.run_command
        return ret

    def _build(self, name: str, proc_def: ProcDef, cache: bool, verbose: bool):
        """Build the conda environment and execute any setup commands."""
        if self._is_personal_env:
            self._env.name = name
        self._env._build(proc_def.root, cache, verbose)

    def _launcher(self, name: str, proc_def: ProcDef):
        if self._env.name == "__defer__":
            self._env.name = name
        return Launcher(self.run_command, name, self._env._env_name(), proc_def)


__all__ = ["Conda"]



from mrp import life_cycle
from mrp import process_def
import click
import json


@click.command()
def cli():
    click.secho("System State:", bold=True)
    click.echo(json.dumps(life_cycle.system_state().asdict(), indent="  "))

    click.secho("Defined Processes:", bold=True)
    click.echo(
        json.dumps(
            {
                name: proc_def.asdict()
                for name, proc_def in process_def.defined_processes.items()
            },
            indent="  ",
        )
    )



from mrp import life_cycle
from mrp.cmd import _autocomplete
import click
import threading
import types


@click.command()
@click.argument("procs", nargs=-1, shell_complete=_autocomplete.running_processes)
@click.option("-t", "--timeout", type=float, default=0)
def cli(*cmd_procs, procs=None, timeout=0):
    procs = procs or []

    # Support procs as *args when using cmd syntax.
    procs += cmd_procs

    wait_procs = set(life_cycle.system_state().procs.keys())
    if procs:
        wait_procs = set(wait_procs) & set(procs)

    ns = types.SimpleNamespace(
        cv=threading.Condition(),
        sat=False,
    )

    def callback(sys_state):
        for proc, info in sys_state.procs.items():
            if proc in wait_procs and info.state != life_cycle.State.STOPPED:
                return
        with ns.cv:
            ns.sat = True
            ns.cv.notify()

    watcher = life_cycle.system_state_watcher(callback)  # noqa: F841

    with ns.cv:
        ns.cv.wait_for(lambda: ns.sat, timeout=timeout or None)



from mrp import life_cycle
from mrp import process_def
import a0
import glob
import os


def union(*fns):
    def impl(ctx, param, incomplete):
        builder = set()
        for fn in fns:
            builder |= set(fn(ctx, param, incomplete))
        return sorted(builder)

    return impl


def intersection(*fns):
    def impl(ctx, param, incomplete):
        builder = None
        for fn in fns:
            if builder is None:
                builder = set(fn(ctx, param, incomplete))
            else:
                builder &= set(fn(ctx, param, incomplete))
        return sorted(builder or [])

    return impl


def conditional(predicate, ontrue, onfalse):
    def impl(ctx, param, incomplete):
        if predicate(ctx, param, incomplete):
            return ontrue(ctx, param, incomplete)
        else:
            return onfalse(ctx, param, incomplete)

    return impl


def defined_processes(ctx, param, incomplete):
    return [
        name
        for name, proc_def in process_def.defined_processes.items()
        if name.startswith(incomplete)
    ]


def running_processes(ctx, param, incomplete):
    state = life_cycle.system_state().asdict()
    return [
        name
        for name, info in state.get("procs", {}).items()
        if info.get("state") == "STARTED" and name.startswith(incomplete)
    ]


def alephzero_topics(protocol):
    def fn(ctx, param, incomplete):
        topics = []
        # TODO(lshamis): Use the a0 pathglob.
        detected = glob.glob(
            os.path.join(a0.env.root(), f"**/*.{protocol}.a0"), recursive=True
        )
        for abspath in detected:
            relpath = os.path.relpath(abspath, a0.env.root())
            topic = relpath[: -len(f".{protocol}.a0")]
            topics.append(topic)

        return [name for name in topics if name.startswith(incomplete)]

    return fn



from setuptools import find_packages
from setuptools import setup


setup(
    name="mrp",
    version="0.1.9",
    author="Leonid Shamis",
    package_dir={"": "src"},
    packages=find_packages(
        where="src",
        include=["*"],
    ),
    python_requires=">=3.7",
    install_requires=[
        "aiodocker>=0.21.0",
        "alephzero>=0.3.11",
        "click>=8.0.3",
        "docker>=5.0.0",
        "psutil>=5.8.0",
        "pyyaml>=6.0",
        "six>=1.16.0",
    ],
    entry_points={
        "console_scripts": ["mrp = mrp.tool:main"],
    },
)



from mrp import life_cycle
import click


@click.command()
def cli():
    state = life_cycle.system_state()
    if not state.procs:
        click.echo("No processes found.")
        return

    name_col_width = max(len(name) for name in state.procs)

    for name, info in state.procs.items():
        suffix = ""
        if info.state == life_cycle.State.STOPPED:
            suffix = f"(stopped code={info.return_code})"
        click.echo(" ".join([name, " " * (name_col_width - len(name)), suffix]))



from mrp.process_def import process
from mrp.runtime.conda import Conda
from mrp.runtime.docker import Docker
from mrp.runtime.host import Host
from mrp.util import NoEscape
from importlib.machinery import SourceFileLoader
import click
import os
import sys


@click.group()
def cli():
    pass


def main(*args):
    try:
        cli(*args)
    except Exception as ex:
        click.echo(ex, err=True)
        sys.exit(1)
    sys.exit(0)


class cmd:
    pass


# Register all commands dynamically into the cli and cmd classes
# to allow execution directly without cli+argv.
#
# TODO(lshamis): Maybe do something similar for runtimes.
#
# For example:
#   mrp.cmd.up(procs=["foo"])
#
#   sys.argv = ["foo"]
#   mrp.main()

this_file_path = os.path.dirname(os.path.realpath(__file__))
cmds_path = os.path.join(this_file_path, "cmd")
for cmd_file in os.listdir(cmds_path):
    if not cmd_file.startswith("_") and cmd_file.endswith(".py"):
        cmd_path = os.path.join(cmds_path, cmd_file)
        cmd_name = cmd_file[: -len(".py")]
        module = SourceFileLoader(cmd_name, cmd_path).load_module()

        setattr(cmd, cmd_name, module.cli.callback)
        cli.add_command(module.cli, cmd_name)

__all__ = ["main", "process", "NoEscape", "Docker", "Conda", "Host", "cmd"]



import mrp
import os
import sys
from importlib.machinery import SourceFileLoader


def main():
    # Default the entrypoint file to the msetup.py in the current directory.
    msetup_path = os.path.abspath("msetup.py")
    # The user may specify another entrypoint with:
    # mrp -f <file> [args...]
    if len(sys.argv) >= 3 and sys.argv[1] == "-f":
        msetup_path = os.path.abspath(sys.argv[2])
        sys.argv = sys.argv[:1] + sys.argv[3:]

    # Show help even if there is no msetup.py.
    if not os.path.exists(msetup_path):
        mrp.main()

    # Run the entrypoint file.
    sys.path.append(os.path.dirname(msetup_path))
    SourceFileLoader("msetup", msetup_path).load_module()


if __name__ == "__main__":
    main()



from mrp import life_cycle
from mrp import util
from mrp.process_def import ProcDef
from mrp.runtime.base import BaseLauncher, BaseRuntime
import asyncio
import os
import pathlib
import signal
import subprocess
import typing


class Launcher(BaseLauncher):
    def __init__(
        self,
        run_command: typing.List[str],
        name: str,
        proc_def: ProcDef,
    ):
        self.run_command = run_command
        self.name = name
        self.proc_def = proc_def

    async def gather_cmd_outputs(self):
        async def log_pipe(logger, pipe):
            while True:
                try:
                    async for line in pipe:
                        logger(line)
                except ValueError:
                    # TODO(lshamis): Can we grab the line in chucks?
                    logger("<[MRP] line length exceeded. skipping>")
                    continue
                else:
                    break

        self.down_task = asyncio.create_task(self.down_watcher(self.handle_down))
        try:
            await asyncio.gather(
                log_pipe(util.stdout_logger(), self.proc.stdout),
                log_pipe(util.stderr_logger(), self.proc.stderr),
                self.log_psutil(),
                self.death_handler(),
                self.down_task,
            )
        except asyncio.exceptions.CancelledError:
            # death_handler cancelled down listener.
            pass

    async def run(self):
        subprocess_env = os.environ.copy()
        subprocess_env.update(self.proc_def.env)

        self.proc = await asyncio.create_subprocess_shell(
            util.shell_join(self.run_command),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            executable="/bin/bash",
            cwd=self.proc_def.root,
            env=subprocess_env,
            start_new_session=True,
        )
        # TODO(lshamis): Handle the case where proc dies before we can query getpgid.
        self.proc_pgrp = os.getpgid(self.proc.pid)

        life_cycle.set_state(self.name, life_cycle.State.STARTED)
        await self.gather_cmd_outputs()

    def get_pid(self):
        return self.proc.pid

    async def death_handler(self):
        await self.proc.wait()
        life_cycle.set_state(
            self.name, life_cycle.State.STOPPED, return_code=self.proc.returncode
        )
        # TODO(lshamis): Restart policy goes here.

        # Release the down listener.
        self.down_task.cancel()

    async def handle_down(self):
        try:
            if self.proc.returncode is None:
                life_cycle.set_state(self.name, life_cycle.State.STOPPING)
                self.proc.send_signal(signal.SIGTERM)

                try:
                    await asyncio.wait_for(self.proc.wait(), timeout=3.0)
                except asyncio.TimeoutError:
                    pass

            # Clean up zombie sub-sub-processes.
            os.killpg(self.proc_pgrp, signal.SIGKILL)
        except:
            pass


class Host(BaseRuntime):
    def __init__(self, run_command, build_commands=None):
        build_commands = build_commands or []

        self.run_command = run_command
        self.build_commands = build_commands

    def asdict(self, root: pathlib.Path):
        ret = {
            "run_command": self.run_command,
        }
        if self.build_commands:
            ret["build_commands"] = self.build_commands
        return ret

    def _build(self, name: str, proc_def: ProcDef, cache: bool, verbose: bool):
        if self.build_commands:
            print(f"Building {name}")
            build_command = "\n".join(
                [util.shell_join(cmd) for cmd in self.build_commands]
            )
            result = subprocess.run(
                build_command,
                shell=True,
                executable="/bin/bash",
                capture_output=not verbose,
            )
            if result.returncode:
                raise RuntimeError(f"Failed to build: {result.stderr}")

    def _launcher(self, name: str, proc_def: ProcDef):
        return Launcher(self.run_command, name, proc_def)


__all__ = ["Host"]
